//
// Copyright (C) 2010  Aleksandar Zlateski <zlateski@mit.edu>
// ----------------------------------------------------------
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

// #ifndef ZI_CACHE_DETAIL_MULTI_INDEX_HPP
// #define ZI_CACHE_DETAIL_MULTI_INDEX_HPP 1

// #include <boost/multi_index_container.hpp>
// #include <boost/multi_index/ordered_index.hpp>
// #include <boost/multi_index/hashed_index.hpp>
// #include <boost/multi_index/member.hpp>

// namespace zi {
// namespace cache {
// namespace multi_index  {

// using namespace boost::multi_index;

// } // namespace multi_array

// using multi_index::multi_index_container;

// } // namespace cache
// } // namespace zi

// #define MULTI_INDEX_MEMBER BOOST_MULTI_INDEX_MEMBER

// #endif

